/**
//Titel:        ClassLoader
//Version:
//Copyright:    Copyright (c) 1999
//@author       J.M.Joller
//Firma:
//Beschreibung:  Beispiel 10_14  ClassLoader Klasse
*/
package Beispiel10_14;
import java.util.Hashtable;
import java.net.*;
import java.io.*;

public class KlasseBeispiel10_14 extends ClassLoader {
    Hashtable cache;
    /** die gelesenen Applets werden in eine Hashtabelle eingetragen
    *   um die Anzahl Lesebefehle zu reduzieren
    */
    URL u;

    public void URLClassLoader(URL u) {
      /** die URLClassLoader Klasse kreiert lediglich eine Hashtabelle
      */
      System.out.println("URLClassLoader Konstruktor");
      this.u = u;
      cache = new Hashtable();
    }
    public synchronized Class loadClass(String  name, boolean resolve)
     throws ClassNotFoundException {
      /** eine Klasse wird zuerst im Cache gesucht
      */
      Class c = (Class) cache.get(name);
      System.out.println("Laden der Klasse "+name);
      /** falls sie dort nicht ist, dann werden die System Klassen durchducht
      */
      if (c == null) {
        try {
          System.out.println("Suchen der Klasse "+name+" im Systempfad.");
          c = findSystemClass(name);
          /** Class Loader findSystemClass Methode zum Suchen und Laden
          *   der Klasse name
          */
        }
        catch (ClassNotFoundException e) {
        }
      }

      // if the class still hasn't been found,
      // load it from the network
      if (c  ==  null) {
        byte b[] = loadClassData(name);
        c = defineClass(b, 0, b.length);
        cache.put(name, c);
      }
      if (resolve) {
        resolveClass(c);
      }

      return  c;

   }  // end loadClass

    private byte[] loadClassData(String name)
     throws ClassNotFoundException {
      System.out.println("Laden der Klasse "+name);
      byte[] b;
      InputStream theClass = null;
      int bfr = 128;

      try {

        URL classURL = new URL(u, name + ".class");

        URLConnection uc = classURL.openConnection();
        uc.setAllowUserInteraction(false);

        try {
          theClass = uc.getInputStream();
        }
        catch (NullPointerException e) {
          System.err.println(e);
          throw new ClassNotFoundException(name + " Input Stream Problem");
        }
        int cl = uc.getContentLength();
        /** Die Lnge der .class Dateien wird vom Server oft bermittelt
        */
        if (cl == -1 ) {
          b = new byte[bfr * 16];
        }
        else {
          b = new byte[cl];
        }

        int bytesread = 0;
        int offset = 0;
        System.out.println("Lesen der Datei in ein Byte Array");
        while (bytesread >= 0) {
          bytesread = theClass.read(b, offset, bfr);
          if (bytesread == -1) break;
          offset += bytesread;
          if (cl == -1 && offset == b.length) { // grow the array
            byte temp[] = new byte[offset * 2];
            System.arraycopy(b, 0, temp, 0, offset);
            b = temp;
          }
          else if (offset > b.length) {
            throw new ClassNotFoundException(name
             + " Fehler beim Lesen der Daten");
          }
        }
        System.out.println("Datei wurde in das Byte Array gelesen");
        /** Optimierung des Byte Arrays (Lnge)
        */
        if (offset < b.length) {
          byte temp[] = new byte[offset];
          System.arraycopy(b, 0, temp, 0, offset);
          b = temp;
        }

        /** wurden alle Bytes empfangen?
        */
        if (cl != -1 && offset != cl) {
          throw new ClassNotFoundException("Es wurden nur " + offset +
            " Bytes von " + name + " empfangen."+
            "\n Erwartet wurden " + cl + " Bytes");
        }
      } // end try
      catch (Exception e) {
       throw new ClassNotFoundException(name + " " + e);
      }
      finally {
        try {
          if (theClass != null) theClass.close();
        }
        catch (IOException e) {
        }

      }

      return b;

    }  // end loadClassData


}  // end URLClassLoader
